/*
 * Created by Brett on 17/11/23.
 * Licensed under GNU General Public License V3.0
 * See LICENSE file for license detail
 */
#include <bf_tokenizer.h>
#include <sstream>
#include <random>
#include <iostream>

struct char_pair
{
    char type;
    size_t count;
};

class characterizer
{
    private:
        std::string program;
        size_t currentIndex = 0;
    public:
        explicit characterizer(std::string program): program(std::move(program))
        {}
        
        inline bool hasNext()
        {
            return currentIndex < program.size();
        }
        
        inline size_t advance()
        {
            return currentIndex++;
        }
        
        inline char next()
        {
            return program[advance()];
        }
        
        inline char_pair fetch()
        {
            size_t start = currentIndex;
            char type = next();
            
            while (peek() == type)
                advance();
            
            return {type, currentIndex - start};
        }
        
        inline char peek()
        {
            return program[currentIndex];
        }
};

std::optional<bf_token> getType(char c)
{
    switch (c)
    {
        case '>':
            return (bf_token::INC_DP);
        case '<':
            return (bf_token::DEC_DP);
        case '+':
            return (bf_token::INC_DV);
        case '-':
            return (bf_token::DEC_DV);
        case '.':
            return (bf_token::PRINT);
        case ',':
            return (bf_token::READ);
        case '[':
            return (bf_token::OPEN);
        case ']':
            return (bf_token::CLOSE);
        default:
            return {};
    }
}

void bf_tokenizer::tokenize(const std::string& program)
{
    characterizer tk{program};
    while (tk.hasNext())
    {
        // TODO: something better
        if (auto type = getType(tk.peek()))
        {
            switch (type.value())
            {
                case bf_token::INC_DP:
                case bf_token::DEC_DP:
                case bf_token::INC_DV:
                case bf_token::DEC_DV:
                    tokens.emplace_back(type.value(), tk.fetch().count);
                    continue;
                case bf_token::PRINT:
                case bf_token::READ:
                case bf_token::OPEN:
                case bf_token::CLOSE:
                    tokens.emplace_back(type.value(), 1);
                    break;
            }
            
        }
        // weird syntax relying on the continue;
        tk.advance();
    }
}

void bf_tokenizer::bf_name()
{
    size_t search_index = 0;
    while (search_index < tokens.size())
    {
        if (tokens[search_index].type == bf_token::OPEN)
        {
            auto name = ++conditionalCount;
            size_t sp = 1;
            size_t search_2 = search_index;
            tokens[search_index].name = name;
            while (++search_2 < tokens.size())
            {
                if (tokens[search_2].type == bf_token::OPEN)
                    sp++;
                if (tokens[search_2].type == bf_token::CLOSE)
                    sp--;
                if (sp == 0)
                {
                    tokens[search_2].offset = static_cast<int64_t>(search_index);
                    break;
                }
            }
            tokens[search_2].name = name;
            tokens[search_index].offset = static_cast<int64_t>(search_2);
        }
        search_index++;
    }
}

inline void tabulate(signed long long v)
{
    for (signed long long i = 0; i < v; i++)
        std::cout << '\t';
}

void bf_tokenizer::print(size_t index)
{
    signed long long sp = 0;
    while (index < tokens.size())
    {
        auto& token = tokens[index++];
        switch (token.type)
        {
            case bf_token::INC_DP:
                tabulate(sp);
                std::cout << "Increase DP " << token.offset << '\n';
                break;
            case bf_token::DEC_DP:
                tabulate(sp);
                std::cout << "Decrease DP " << token.offset << '\n';
                break;
            case bf_token::INC_DV:
                tabulate(sp);
                std::cout << "Increase DV " << token.offset << '\n';
                break;
            case bf_token::DEC_DV:
                tabulate(sp);
                std::cout << "Decrease DV " << token.offset << '\n';
                break;
            case bf_token::PRINT:
                tabulate(sp);
                std::cout << "Print\n";
                break;
            case bf_token::READ:
                tabulate(sp);
                std::cout << "Read\n";
                break;
            case bf_token::OPEN:
                tabulate(sp);
                std::cout << "If(";
                if (token.name.has_value())
                    std::cout << token.name.value() << "\n";
                else
                    std::cout << "UNNAMED" << "\n";
                sp++;
                break;
            case bf_token::CLOSE:
                sp--;
                tabulate(sp);
                if (token.name.has_value())
                    std::cout << token.name.value();
                else
                    std::cout << "UNNAMED";
                std::cout << ")\n";
                break;
        }
    }
}

std::string generateName()
{
    std::stringstream ss;
    ss << std::hex;
    static std::random_device rd;
    static std::seed_seq seed{rd(), rd(), rd(), rd()};
    static std::mt19937_64 gen(seed);
    static std::uniform_int_distribution<int> dis(0, 15);
    
    for (int i = 0; i < 24; i++)
        ss << dis(gen);
    
    return ss.str();
}